@@ -17,8 +17,10 @@
 #' @export
 #'
 #' @examples
+#' \donttest{
 #' vst_out <- vst(pbmc)
 #' y_smooth <- smooth_via_pca(vst_out$y, do_plot = TRUE)
+#' }
 #'
 smooth_via_pca <- function(x, elbow_th = 0.025, dims_use = NULL, max_pc = 100, do_plot = FALSE,
                            scale. = FALSE) {
@@ -60,18 +62,37 @@ smooth_via_pca <- function(x, elbow_th = 0.025, dims_use = NULL, max_pc = 100, d
 #' @param cell_attr Provide cell meta data holding latent data info
 #' @param do_round Round the result to integers
 #' @param do_pos Set negative values in the result to zero
-#' @param show_progress Whether to print progress bar
+#' @param verbosity An integer specifying whether to show only messages (1), messages and progress bars (2) or nothing (0) while the function is running; default is 2
+#' @param verbose Deprecated; use verbosity instead
+#' @param show_progress Deprecated; use verbosity instead
 #'
 #' @return Corrected data as UMI counts
 #'
 #' @export
 #'
 #' @examples
+#' \donttest{
 #' vst_out <- vst(pbmc, return_cell_attr = TRUE)
 #' umi_corrected <- correct(vst_out)
+#' }
 #'
-correct <- function(x, data = 'y', cell_attr = x$cell_attr, do_round = TRUE, do_pos = TRUE,
-                    show_progress = TRUE) {
+correct <- function(x, data = 'y', cell_attr = x$cell_attr, do_round = TRUE,
+                    do_pos = TRUE, verbosity = 2, verbose = TRUE, show_progress = TRUE) {
+  # Take care of deprecated arguments
+  args_passed <- names(sapply(match.call(), deparse))[-1]
+  if ('verbose' %in% args_passed) {
+    warning("The 'verbose' argument is deprecated as of v0.3. Use 'verbosity' instead.", immediate. = TRUE)
+    verbosity <- as.numeric(verbose)
+  }
+  if ('show_progress' %in% args_passed) {
+    warning("The 'show_progress' argument is deprecated as of v0.3. Use 'verbosity' instead.", immediate. = TRUE)
+    if (show_progress) {
+      verbosity <- 2
+    } else {
+      verbosity <- min(verbosity, 1)
+    }
+  }
+
   if (is.character(data)) {
     data <- x[[data]]
   }
@@ -83,8 +104,10 @@ correct <- function(x, data = 'y', cell_attr = x$cell_attr, do_round = TRUE, do_
   bin_size <- x$arguments$bin_size
   bin_ind <- ceiling(x = 1:length(x = genes) / bin_size)
   max_bin <- max(bin_ind)
-  if (show_progress) {
+  if (verbosity > 0) {
     message('Computing corrected count matrix for ', length(genes), ' genes')
+  }
+  if (verbosity > 1) {
     pb <- txtProgressBar(min = 0, max = max_bin, style = 3)
   }
   corrected_data <- matrix(NA_real_, length(genes), nrow(regressor_data), dimnames = list(genes, rownames(regressor_data)))
@@ -96,11 +119,11 @@ correct <- function(x, data = 'y', cell_attr = x$cell_attr, do_round = TRUE, do_
     mu <- exp(tcrossprod(coefs, regressor_data))
     variance <- mu + mu^2 / theta
     corrected_data[genes_bin, ] <- mu + pearson_residual * sqrt(variance)
-    if (show_progress) {
+    if (verbosity > 1) {
       setTxtProgressBar(pb, i)
     }
   }
-  if (show_progress) {
+  if (verbosity > 1) {
     close(pb)
   }
 
@@ -122,7 +145,9 @@ correct <- function(x, data = 'y', cell_attr = x$cell_attr, do_round = TRUE, do_
 #' @param x A list that provides model parameters and optionally meta data; use output of vst function
 #' @param umi The count matrix
 #' @param cell_attr Provide cell meta data holding latent data info
-#' @param show_progress Whether to print progress bar
+#' @param verbosity An integer specifying whether to show only messages (1), messages and progress bars (2) or nothing (0) while the function is running; default is 2
+#' @param verbose Deprecated; use verbosity instead
+#' @param show_progress Deprecated; use verbosity instead
 #'
 #' @return Corrected data as UMI counts
 #'
@@ -131,10 +156,28 @@ correct <- function(x, data = 'y', cell_attr = x$cell_attr, do_round = TRUE, do_
 #' @export
 #'
 #' @examples
+#' \donttest{
 #' vst_out <- vst(pbmc, return_cell_attr = TRUE)
 #' umi_corrected <- correct_counts(vst_out, pbmc)
+#' }
 #'
-correct_counts <- function(x, umi, cell_attr = x$cell_attr, show_progress = TRUE) {
+correct_counts <- function(x, umi, cell_attr = x$cell_attr, verbosity = 2,
+                           verbose = TRUE, show_progress = TRUE) {
+  # Take care of deprecated arguments
+  args_passed <- names(sapply(match.call(), deparse))[-1]
+  if ('verbose' %in% args_passed) {
+    warning("The 'verbose' argument is deprecated as of v0.3. Use 'verbosity' instead.", immediate. = TRUE)
+    verbosity <- as.numeric(verbose)
+  }
+  if ('show_progress' %in% args_passed) {
+    warning("The 'show_progress' argument is deprecated as of v0.3. Use 'verbosity' instead.", immediate. = TRUE)
+    if (show_progress) {
+      verbosity <- 2
+    } else {
+      verbosity <- min(verbosity, 1)
+    }
+  }
+
   regressor_data_orig <- model.matrix(as.formula(gsub('^y', '', x$model_str)), cell_attr)
   # when correcting, set all latent variables to median values
   cell_attr[, x$arguments$latent_var] <- apply(cell_attr[, x$arguments$latent_var, drop=FALSE], 2, function(x) rep(median(x), length(x)))
@@ -144,15 +187,17 @@ correct_counts <- function(x, umi, cell_attr = x$cell_attr, show_progress = TRUE
   bin_size <- x$arguments$bin_size
   bin_ind <- ceiling(x = 1:length(x = genes) / bin_size)
   max_bin <- max(bin_ind)
-  if (show_progress) {
+  if (verbosity > 0) {
     message('Computing corrected UMI count matrix')
+  }
+  if (verbosity > 1) {
     pb <- txtProgressBar(min = 0, max = max_bin, style = 3)
   }
   #corrected_data <- matrix(NA_real_, length(genes), nrow(regressor_data), dimnames = list(genes, rownames(regressor_data)))
   corrected_data <- list()
   for (i in 1:max_bin) {
     genes_bin <- genes[bin_ind == i]
-    coefs <- x$model_pars_fit[genes_bin, -1]
+    coefs <- x$model_pars_fit[genes_bin, -1, drop=FALSE]
     theta <- x$model_pars_fit[genes_bin, 1]
     # get pearson residuals
     mu <- exp(tcrossprod(coefs, regressor_data_orig))
@@ -166,11 +211,11 @@ correct_counts <- function(x, umi, cell_attr = x$cell_attr, show_progress = TRUE
     y.res <- round(y.res, 0)
     y.res[y.res < 0] <- 0
     corrected_data[[length(corrected_data) + 1]] <- as(y.res, Class = 'dgCMatrix')
-    if (show_progress) {
+    if (verbosity > 1) {
       setTxtProgressBar(pb, i)
     }
   }
-  if (show_progress) {
+  if (verbosity > 1) {
     close(pb)
   }
   corrected_data <- do.call(what = rbind, args = corrected_data)
